﻿using UnityEngine;
using UnityEngine.UI;
using System.Collections;

/*! \mainpage Index Page
 * 
 * \section Short Explanation
 * The source code is located in the \ref src directory.
 * 
 * The functionality of the executable includes loading and visualizing Tensorflow weights and featuremap activations from a custom Json format that has to be converted with a python script from original .ckpt binary files.
 * 
 * This can be very useful to teach and explain the flow of computation of a CNN, as well as to inspect the change of weights and activations over the course of a training run, because the loading and switching between epochs is supported as well.
 * 
 * After loading the training sample by typing its number into the text field and pressing the "load" button, a number of parameters can be changed, most notably the epoch and the line width, the latter enabling an unobstructed view at all feature maps when set to zero, as seen in the image below.
 * 
 * Navigation: Alt + Left Mouse for rotation, Middle for panning and Right for dollying. 
 * 
 */

/// <summary>
/// Acts as an interface between the GUI and other classes.
/// </summary>
public class GuiManager : MonoBehaviour
{

    public float ConvfullResHeight = 5.0f;

    public Text predictedLabel;

    public Material lineMaterial;

    public Material pixelMaterial;

    private float expansionCenter = 0.5f;
    private float expansionWidth = 0.2f;
    private int epoch = 0;

    // Use this for initialization
    void Start()
    {

    }

    // Update is called once per frame
    void Update()
    {
        //SetExpansionCenter(0.5f+Mathf.Sin(Time.fixedTime*4)/2f);
    }

    private void OnValidate()
    {
        GlobalManager.Instance.SetFullResHeight(ConvfullResHeight);
    }

    /// <summary>
    /// Logs Layer names for Debugging
    /// </summary>
    public void LogLayers()
    {
        foreach(Layer l in GlobalManager.Instance.GetAllLayersOrdered())
        {
            Debug.Log(l.name);
        }
    }

    /// <summary>
    /// Sets the center of the expanded layer range.
    /// </summary>
    /// <param name="value"></param>
    public void SetExpansionCenter(float value)
    {
        expansionCenter = value;
        GlobalManager.Instance.SetExpansion(expansionCenter, expansionWidth);
    }

    /// <summary>
    /// Sets the width of the expanded layer range.
    /// </summary>
    /// <param name="value"></param>
    public void SetExpansionWidth(float value)
    {
        expansionWidth = value;
        GlobalManager.Instance.SetExpansion(expansionCenter, expansionWidth);
    }

    /// <summary>
    /// Sets the epoch to be loaded.
    /// </summary>
    /// <param name="value"></param>
    public void SetEpoch(float value)
    {
        epoch = (int)value;
        GlobalManager.Instance.SetEpoch(epoch);
        predictedLabel.text = GlobalManager.Instance.predPerEpoch[epoch];
    }

    /// <summary>
    /// Set the line width of the edges.
    /// </summary>
    /// <param name="value"></param>
    public void SetLineWidth(float value)
    {
        lineMaterial.SetFloat("_width", value);
    }

    /// <summary>
    /// Sets the brightness of all nodes and edges (color contrast when rendering red/blue mapping)
    /// </summary>
    /// <param name="value"></param>
    public void SetBrightness(float value)
    {
        GlobalManager.Instance.SetBrightness(value);
    }

    public void SetFullres(bool value)
    {
        GlobalManager.Instance.SetFullResDisplay(value);
    }

    public void SetSquarePixels(bool value)
    {
        pixelMaterial.SetInt("_squarePixels", value ? 1 : 0);
    }

    public void SetBWPixels(bool value)
    {
        pixelMaterial.SetInt("_useBlueRedCmap", value ? 0 : 1);
    }

    /// <summary>
    /// Load the currently selected Testsample epochs
    /// </summary>
    public void Load()
    {
        TFLoader loader = GameObject.FindObjectOfType<TFLoader>();
        loader.Load();
    }

}
