﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

/// <summary>
/// Represents a Fully Connected Layer of a CNN.
/// </summary>
public class FCLayer : Layer
{
    public int temp_maxHRuns = 1;
    public int temp_maxIRuns = 1;
    public int temp_maxJRuns = 1;
    public int temp_maxKRuns = 1;

    public int reducedDepth = 16;

    public int fullDepth = 1024;
    private int oldFullDepth;

    private int oldReducedDepth;

    private Vector3Int inputResolution;

    [Range(0.0f, 1.0f)]
    public float filterSpread;

    [Range(0.0f, 0.1f)]
    public float filterSpacing;

    [Range(0.0f, 2.0f)]
    public float lineCircleGrid;

    [Range(0.0f, 1.0f)]
    public float edgeBundle;

    [Range(0.0f, 1.0f)]
    public float collapseInput;

    [Range(0.0f, 1.0f)]
    public float lineXToZ;

    public float nodeSize;
    public bool showOriginalDepth;

    private List<Vector3> _nodePositions;

    private Dictionary<int, Array> _tensorPerEpoch = new Dictionary<int, Array>();
    private int[] tensorShape = new int[4];

    private Dictionary<int, Array> _activationTensorPerEpoch = new Dictionary<int, Array>();
    private int[] activationTensorShape = new int[4];

    public FCLayer()
    {

    }

    public override void SetExpansion(float expansion)
    {
        collapseInput = 1-expansion;
        base.SetExpansion(expansion);
    }

    protected override void SubUpdate()
    {
        edgeBundle = Mathf.Sin(Time.fixedTime) / 2.0f + 0.5f;
        lineCircleGrid = Mathf.Sin(Time.fixedTime) + 1.0f;
    }

    /// <summary>
    /// Checks if parameters have been updated and reinitializes necessary data.
    /// </summary>
    protected override void UpdateForParams()
    {
        Vector3Int newRes = inputResolution;
        if (_inputLayer) {
            newRes = _inputLayer.GetOutputShape();
        }

        if(reducedDepth != oldReducedDepth ||
            newRes != inputResolution ||
            _initialized == false ||
            _nodePositions == null ||
            oldFullDepth != fullDepth)
        {
            inputResolution = newRes;
            InitNodes();
            oldReducedDepth = reducedDepth;
            oldFullDepth = fullDepth;
        }
    }

    /// <summary>
    /// Initialize data representing nodes of the layer.
    /// </summary>
    private void InitNodes()
    {
        Vector3[] filterPositions = GetInterpolatedNodePositions();

        _nodePositions = new List<Vector3>();
        for (int i = 0; i < reducedDepth; i++)
        {
            Vector3 node = filterPositions[i];
            _nodePositions.Add(node);
        }
    }

    /// <summary>
    /// Update data representing nodes of the layer.
    /// </summary>
    private void UpdateNodes()
    {
        Vector3[] filterPositions = GetInterpolatedNodePositions();

        for (int i=0; i<_nodePositions.Count; i++)
        {
            _nodePositions[i] = filterPositions[i];
        }
    }

    /// <summary>
    /// Calculates the position of the nodes according to parameters.
    /// </summary>
    /// <returns></returns>
    private Vector3[] GetInterpolatedNodePositions()
    {
        int reducedDepthGridShape = Mathf.CeilToInt(Mathf.Sqrt(reducedDepth));
        Vector3[] gridPositions = GridShape.ScaledUnitGrid(reducedDepthGridShape, reducedDepthGridShape, new Vector3(0, 0, 0), filterSpread);
        Vector3[] linePositionsX = LineShape.ScaledUnitLine(reducedDepth, new Vector3(0, 0, 0), new Vector3(1, 0, 0),filterSpread * (reducedDepthGridShape / (float)reducedDepth));
        Vector3[] linePositionsZ = LineShape.ScaledUnitLine(reducedDepth, new Vector3(0, 0, 0), new Vector3(0, 0, 1), filterSpread * (reducedDepthGridShape / (float)reducedDepth));
        Vector3[] linePositions = Shape.InterpolateShapes(linePositionsX, linePositionsZ, lineXToZ);
        Vector3[] circlePositions = CircleShape.ScaledUnitCircle(reducedDepth, new Vector3(0, 0, 0), filterSpread);

        Vector3[] filterPositions = null;

        if (lineCircleGrid < 1.0f)
        {
            filterPositions = Shape.InterpolateShapes(linePositions, circlePositions, lineCircleGrid);
        }
        else if (lineCircleGrid <= 2.0f)
        {
            filterPositions = Shape.InterpolateShapes(circlePositions, gridPositions, lineCircleGrid - 1.0f);
        }
        return filterPositions;
    }

    /// <summary>
    /// Returns a List of Shapes representing the pixels of the feature maps.
    /// </summary>
    /// <returns></returns>
    protected override List<Shape> GetPointShapes()
    {
        UpdateNodes();

        List<Shape> pixelGrids = new List<Shape>();
        for(int i=0; i<_nodePositions.Count; i++)
        {
            GridShape dummyGrid = new GridShape(_nodePositions[i], DrawMode.Dot, new Vector2Int(1, 1), new Vector2(0.0f, 0.0f));
            pixelGrids.Add(dummyGrid);
        }
        return pixelGrids;
    }

    /// <summary>
    /// Calculates and returns the positions of the line start points (for the CalcMesh() function)
    /// </summary>
    /// <param name="convShape"></param>
    /// <param name="outputShape"></param>
    /// <param name="theoreticalOutputShape"></param>
    /// <param name="stride"></param>
    /// <param name="allCalcs"></param>
    /// <returns></returns>
    public override List<List<Shape>> GetLineStartShapes(Vector2Int convShape, Vector2Int outputShape, Vector2 theoreticalOutputShape, Vector2Int stride, float allCalcs)
    {
        List<List<Shape>> list = new List<List<Shape>>();
        list.Add(GetPointShapes());
        return list;
    }


    override public void CalcMesh()
    {
        if(_input == null)
        {
            base.CalcMesh();
            return;
        }

        _mesh.subMeshCount = 2;


        //POINTS
        List<Vector3> verts = new List<Vector3>();
        List<Color> cols = new List<Color>();
        List<int> inds = new List<int>();

        Vector3 posDiff = new Vector3(0, 0, -zOffset);
        Vector3 zPos = new Vector3(0, 0, ZPosition());

        AddNodes(verts, inds);
        for (int i = 0; i < verts.Count; i++)
        {
            if (_activationTensorPerEpoch.ContainsKey(epoch))
            {

                int[] index = GetMultiDimIndices(activationTensorShape, i);

                Array activationTensor = _activationTensorPerEpoch[epoch];
                float tensorVal = (float)activationTensor.GetValue(index) * pointBrightness;
                cols.Add(new Color(tensorVal, tensorVal, tensorVal, 1f));

                //cols.Add(Color.white);
            }
            else
            {
                cols.Add(Color.black);
            }
        }



        if (showOriginalDepth)
        {
            AddFullResNodes(verts, inds);
        }
        int diff = verts.Count - cols.Count;
        for (int i = 0; i < diff; i++)
        {
            cols.Add(Color.black);
        }

        //LINES
        Vector2Int inputShape = new Vector2Int(1, 1);

        if(collapseInput != 1.0f)
        {
            inputShape = new Vector2Int(_inputLayer.GetOutputShape().x, _inputLayer.GetOutputShape().y);
        }

        //Input Layer Points, each inner List only contains one point
        List<List<Shape>> inputFilterPoints = _inputLayer.GetLineStartShapes(inputShape, new Vector2Int(1, 1), new Vector2(1, 1), new Vector2Int(1, 1), 0.0f);

        List<int> lineInds = new List<int>();

        if (edgeBundle == 1f)
        {
            AddFullyBundledLines(verts, lineInds, inputFilterPoints);
            diff = verts.Count - cols.Count;
            for (int i = 0; i < diff; i++)
            {
                cols.Add(Color.black);
            }
        }
        else
        {

            //for each of this Layers Points
            for (int h = 0; h < _nodePositions.Count; h++)
            {
                //line endpoint
                Vector3 end_vert = _nodePositions[h];
                verts.Add(end_vert + zPos);
                cols.Add(Color.black);

                int end_ind = verts.Count - 1;

                Vector3 edgeBundleCenter = GetEdgeBundleCenter(end_vert, edgeBundle);
                int bundle_center_ind = 0;
                if (edgeBundle > 0)
                {

                    verts.Add(edgeBundleCenter);
                    cols.Add(Color.black);

                    bundle_center_ind = verts.Count - 1;
                }

                //for each input feature map
                for (int i = 0; i < inputFilterPoints.Count; i++)
                {
                    //for each input conv grid
                    List<Shape> featureMapGrids = inputFilterPoints[i];
                    for (int j = 0; j < featureMapGrids.Count; j++)
                    {
                        GridShape gridShape = (GridShape)featureMapGrids[j];
                        //scale shape spacing by collapse input
                        gridShape.spacing *= (1.0f - collapseInput);
                        Vector3[] start_verts = gridShape.GetVertices(true);

                        //for each conv point
                        for (int k = 0; k < start_verts.Length; k++)
                        {
                            lineInds.Add(end_ind);
                            if (edgeBundle > 0)
                            {
                                lineInds.Add(bundle_center_ind);
                                lineInds.Add(bundle_center_ind);
                            }

                            verts.Add(start_verts[k] + zPos + posDiff);
                            if (_tensorPerEpoch.ContainsKey(epoch))
                            {
                                Array tensor = _tensorPerEpoch[epoch];

                                int[] index = { i*k, h };
                                if (_inputLayer.GetType().Equals(typeof(FCLayer)))
                                {
                                    index[0] = j;
                                }

                                float tensorVal = (float)tensor.GetValue(index) * 5f;
                                cols.Add(new Color(tensorVal, tensorVal, tensorVal, 1f));
                            }
                            else
                            {
                                cols.Add(Color.black);
                            }

                            lineInds.Add(verts.Count - 1);
                        }
                    }
                }
            }
        }


        _mesh.SetVertices(verts);
        _mesh.SetColors(cols);
        _mesh.SetIndices(inds.ToArray(), MeshTopology.Points, 0);
        _mesh.SetIndices(lineInds.ToArray(), MeshTopology.Lines, 1);

        Debug.Log("vertcount: " + verts.Count);
    }

    /// <summary>
    /// As we require much less vertices and line polygons when the edge bundling is turned up to 1.0, the mesh is calculated differently (in this function).
    /// </summary>
    /// <param name="verts"></param>
    /// <param name="lineInds"></param>
    /// <param name="inputFilterPoints"></param>
    void AddFullyBundledLines(List<Vector3> verts, List<int> lineInds, List<List<Shape>> inputFilterPoints)
    {
        Vector3 posDiff = new Vector3(0, 0, -zOffset);
        Vector3 zPos = new Vector3(0, 0, ZPosition());

        Vector3 end_vert0 = _nodePositions[0];
        Vector3 edgeBundleCenter = GetEdgeBundleCenter(end_vert0, 1f);
        verts.Add(edgeBundleCenter);
        int bundle_center_ind = verts.Count - 1;

        //for each of this Layers Points
        for (int h = 0; h < _nodePositions.Count; h++)
        {
            //line endpoint
            Vector3 end_vert = _nodePositions[h];
            verts.Add(end_vert + zPos);

            int end_ind = verts.Count - 1;

            lineInds.Add(end_ind);
            lineInds.Add(bundle_center_ind);
        }

        //for each input feature map
        for (int i = 0; i < inputFilterPoints.Count; i++)
        {
            //for each input conv grid
            List<Shape> featureMapGrids = inputFilterPoints[i];
            for (int j = 0; j < featureMapGrids.Count; j++)
            {
                GridShape gridShape = (GridShape)featureMapGrids[j];
                //scale shape spacing by collapse input
                gridShape.spacing *= (1.0f - collapseInput);
                Vector3[] start_verts = gridShape.GetVertices(true);

                //for each conv point
                for (int k = 0; k < start_verts.Length; k++)
                {

                    lineInds.Add(bundle_center_ind);

                    verts.Add(start_verts[k] + zPos + posDiff);
                    lineInds.Add(verts.Count - 1);
                }
            }
        }
    }

    /// <summary>
    /// Adds nodes visualizing the full, unreduced resolution of the layer.
    /// </summary>
    /// <param name="verts"></param>
    /// <param name="inds"></param>
    private void AddFullResNodes(List<Vector3> verts, List<int> inds)
    {
        Vector3[] vertsToAdd = null;

        if (lineCircleGrid < 1.0f)
        {
            Vector3[] lineVerts = fullResLineVerts();
            Vector3[] circleVerts = fullResCircleVerts();
            vertsToAdd = interpolateVectors(lineVerts, circleVerts, lineCircleGrid);
        }
        else if (lineCircleGrid <= 2.0f)
        {
            Vector3[] circleVerts = fullResCircleVerts();
            Vector3[] gridVerts = fullResGridVerts();
            vertsToAdd = interpolateVectors(circleVerts, gridVerts, lineCircleGrid - 1.0f);
        }

        for(int i=0; i<vertsToAdd.Length; i++)
        {
            verts.Add(vertsToAdd[i]);
            inds.Add(verts.Count - 1);
        }
    }

    /// <summary>
    /// helper to interpolate two vectors.
    /// </summary>
    /// <param name="a"></param>
    /// <param name="b"></param>
    /// <param name="alpha"></param>
    /// <returns></returns>
    private Vector3[] interpolateVectors(Vector3[] a, Vector3[] b, float alpha)
    {
        for(int i=0; i<Mathf.Min(a.Length, b.Length); i++)
        {
            a[i] = a[i] * (1.0f - alpha) + b[i] * alpha;
        }
        return a;
    }

    /// <summary>
    /// Returns the vertices of the line used for the full res visualization
    /// </summary>
    /// <returns></returns>
    private Vector3[] fullResLineVerts()
    {
        int reducedDepthGridShape = Mathf.CeilToInt(Mathf.Sqrt(reducedDepth));

        LineShape line = new LineShape(CenterPosition(), DrawMode.Dot, fullDepth, filterSpread * (reducedDepthGridShape / (float)reducedDepth));
        return line.GetVertices(false);
    }

    /// <summary>
    /// Returns the vertices of the circles used for the full res visualization
    /// </summary>
    /// <returns></returns>
    private Vector3[] fullResCircleVerts()
    {
        List<Vector3> outVerts = new List<Vector3>();
        int circleNum = Mathf.CeilToInt( fullDepth / (float)reducedDepth);

        for(int i=0; i<circleNum; i++)
        {
            CircleShape circle = new CircleShape(CenterPosition(), DrawMode.Dot, reducedDepth, filterSpread * 1.3f + i*2.0f*(filterSpread / 10.0f));
            outVerts.AddRange(circle.GetVertices(false));
        }


        return outVerts.ToArray();
    }

    /// <summary>
    /// Returns the vertices of the grid used for the full res visualization
    /// </summary>
    /// <returns></returns>
    private Vector3[] fullResGridVerts()
    {
        int root = Mathf.CeilToInt(Mathf.Sqrt(fullDepth));

        GridShape grid = new GridShape(CenterPosition(), DrawMode.Dot, new Vector2Int(root, root), new Vector2(filterSpread, filterSpread));
        return grid.GetVertices(false);
    }

    public override Vector3Int GetOutputShape()
    {
        return new Vector3Int(1, 1, reducedDepth);
    }

    public void SetTensorForEpoch(Array tensor, int epoch)
    {
        _tensorPerEpoch.Add(epoch, tensor);
    }

    public Array GetTensorForEpoch(int epoch)
    {
        return _tensorPerEpoch[epoch];
    }

    public void SetTensorShape(int[] tensorShape)
    {
        this.tensorShape = tensorShape;
    }

    public int[] GetTensorShape()
    {
        return tensorShape;
    }

    public void SetActivationTensorForEpoch(Array tensor, int epoch)
    {
        _activationTensorPerEpoch.Add(epoch, tensor);
    }

    public Array GetActivationTensorForEpoch(int epoch)
    {
        return _activationTensorPerEpoch[epoch];
    }

    public void SetActivationTensorShape(int[] tensorShape)
    {
        this.activationTensorShape = tensorShape;
    }

    public int[] GetActivationTensorShape()
    {
        return activationTensorShape;
    }
}
