﻿using UnityEngine;
using System.Collections.Generic;

public class FeatureMap
{
    public Vector3 position;
    private Vector3 outputPosition;
    public Vector2Int shape;
    public Vector2Int filterShape;
    public Vector2Int outputShape;
    private Vector2 theoreticalOutputShape;

    public Vector2Int stride = new Vector2Int(1, 1);

    private GridShape _pixelGrid;
    private Shape _filterGrid;
    private GridShape _outputGrid;
    private List<Shape> _allCalcFilterGrids;

    //TODO: make 2 dimensional
    public Vector2 spacing;

    public FeatureMap(Vector3 position, Vector2Int shape, Vector2Int filterShape, Vector2Int outputShape, float spacing)
    {
        this.position = position;
        this.outputPosition = position;
        this.shape = shape;
        this.outputShape = outputShape;
        this.filterShape = filterShape;
        this.spacing = new Vector2(spacing, spacing);

        InitGrids();
    }

    public GridShape GetPixelGrid()
    {
        return _pixelGrid;
    }

    public List<Shape> GetFilterGrids(Vector2Int outputShape, Vector2 theoreticalOutputShape, Vector2Int stride, float allCalcs)
    {
        //check if requested outputshape is same as existing, reinit allcalgrids if not
        if(outputShape != this.outputShape
            ||  stride != this.stride
            || this.theoreticalOutputShape != theoreticalOutputShape
            && outputShape != new Vector2Int(0, 0))
        {
            this.outputShape = outputShape;
            this.theoreticalOutputShape = theoreticalOutputShape;
            this.outputPosition = position + GetOutputGridOffset(theoreticalOutputShape, outputShape);
            this.stride = stride;
            InitGrids();
        }

        if (allCalcs == 0)
        {
            List<Shape> filterGrids = new List<Shape>();
            filterGrids.Add(_filterGrid);
            return filterGrids;
        } else
        {
            List<Shape> filterGrids = new List<Shape>();
            foreach(GridShape gr in _allCalcFilterGrids)
            {

                gr.spacing /= (shape.x - 1) / (float)(filterShape.x - 1);
                GridShape interpolated = gr.InterpolatedGrid(((GridShape)_filterGrid), 1.0f - allCalcs);
                filterGrids.Add(interpolated);
            }
            return filterGrids;
        }
    }


    private Vector3 GetOutputGridOffset(Vector2 theoreticalOutputShape, Vector2Int outputShape)
    {
        Vector2 safeOffset = new Vector2(0, 0);
        if (theoreticalOutputShape != outputShape)
        {
            if (outputShape.x > 1 && outputShape.y > 1)
                safeOffset = (spacing) * 0.5f;
        }
        return new Vector3(safeOffset.x, safeOffset.y, 0);
    }

    public Shape GetInputGrid(float allCalcs)
    {
        if (allCalcs == 0)
        {

            return new GridShape(position, DrawMode.ConnectedDot, shape, new Vector2(0, 0));
        }
        else if (allCalcs == 1.0f)
        {
            return _pixelGrid;
        }
        else
        {
            GridShape degenerate = new GridShape(position, DrawMode.ConnectedDot, shape, new Vector2(0, 0));
            GridShape interpolated = degenerate.InterpolatedGrid(_pixelGrid, allCalcs);

            return interpolated;
        }
    }

    private void InitGrids()
    {
        _pixelGrid = new GridShape(position, DrawMode.Dot, shape, spacing);
        _outputGrid = new GridShape(outputPosition, DrawMode.Dot, shape, spacing);
        _allCalcFilterGrids = new List<Shape>();

        Vector2 safeSpacing = new Vector2(0.0f, 0.0f);
        if(filterShape.x > 1)
        {
            safeSpacing = (shape.x - 1) / (float)(filterShape.x - 1) * spacing;
        }

        _filterGrid = new GridShape(outputPosition, DrawMode.ConnectedDot, filterShape, safeSpacing); 

        Vector3[] allCalcPositions;
      
        if(outputShape == shape) //means stride == 1
        {
            allCalcPositions = _pixelGrid.GetVertices(true);
        } else
        {
            _outputGrid = new GridShape(outputPosition, DrawMode.ConnectedDot, outputShape, spacing * stride.x);
            allCalcPositions = _outputGrid.GetVertices(true);
        }

        for (int i = 0; i < allCalcPositions.Length; i++)
        {
            _allCalcFilterGrids.Add(new GridShape(allCalcPositions[i], DrawMode.ConnectedDot, filterShape, safeSpacing));
        }

    }

    private void UpdateGrids()
    {
        _pixelGrid.position = position;
        _pixelGrid.resolution = shape;
        _pixelGrid.spacing = spacing;

        _outputGrid.position = outputPosition;
        _outputGrid.resolution = outputShape;
        _outputGrid.spacing = spacing * stride;


        _filterGrid.position = outputPosition;

        Vector2 safeSpacing = new Vector2(0.0f, 0.0f);
        if (filterShape.x > 1)
        {
            safeSpacing = (shape.x - 1) / (float)(filterShape.x - 1) * spacing;
        }

        ((GridShape)_filterGrid).spacing = safeSpacing;

        Vector3[] allCalcPositions;

        if (outputShape == shape) //means stride == 1
        {
            allCalcPositions = _pixelGrid.GetVertices(true);
        }
        else
        {
            _outputGrid.position = outputPosition;
            _outputGrid.resolution = outputShape;
            _outputGrid.spacing = safeSpacing;

            allCalcPositions = _outputGrid.GetVertices(true);
        }

        for (int i = 0; i < allCalcPositions.Length; i++)
        {
            ((GridShape)_allCalcFilterGrids[i]).position = allCalcPositions[i];
            ((GridShape)_allCalcFilterGrids[i]).resolution = filterShape;
            ((GridShape)_allCalcFilterGrids[i]).spacing = safeSpacing;
        }
    }

    public void UpdateValues(Vector3 position, Vector2Int shape, Vector2Int filterShape, Vector2Int outputShape, float spacing)
    {
        this.position = position;
        this.outputPosition = position;
        this.spacing = new Vector2(spacing, spacing);

        bool reinit = false;
        if (this.shape != shape || this.filterShape != filterShape || this.outputShape != outputShape) reinit = true;

        this.shape = shape;
        this.filterShape = filterShape;
        this.outputShape = outputShape;

        if (reinit) InitGrids();
        else UpdateGrids();
    }

    public static Vector2Int GetFeatureMapShapeFromInput(Vector2Int inputShape, Vector2Int filterShape, Vector2Int inputStride, Vector2Int padding)
    {
        Vector2 featureMapDims = (inputShape - filterShape + new Vector2(2f, 2f) * padding) / (Vector2) inputStride + new Vector2(1f, 1f);
        Vector2Int intFeatureMapDims = Vector2Int.FloorToInt(featureMapDims);

        return intFeatureMapDims;
    }

    public static Vector2 GetTheoreticalFloatFeatureMapShapeFromInput(Vector2Int inputShape, Vector2Int filterShape, Vector2Int inputStride, Vector2Int padding)
    {
        Vector2 featureMapDims = (inputShape - filterShape + new Vector2(2f, 2f) * padding) / inputStride + new Vector2(1f, 1f);

        return featureMapDims;
    }
}