﻿using UnityEngine;
using System.Collections;
using System.Collections.Generic;
using System;

public class ImageLayer : Layer
{
    public Vector2Int fullResolution = new Vector2Int(224, 224);
    private Vector2Int oldFullResolution;

    public int depth = 3;
    private int oldDepth;

    public bool rgb = true;

    public Vector2Int reducedResolution = new Vector2Int(11, 11);
    private Vector2Int oldreducedResolution;

    private Vector2Int currentConvShape = new Vector2Int(1, 1);

    [Range(0.0f, 1.0f)]
    public float spread = 1.0f;

    [Range(0.0f, 0.1f)]
    public float pixelSpacing = 0.05f;

    [Range(0.0f, 1.0f)]
    public float xToZ = 0.0f;

    [Range(0.1f, 5.0f)]
    public float fullresOffset = 1.0f;

    public float nodeSize;
    public bool showOriginalResolution = false;

    private List<FeatureMap> _featureMaps;

    private GridShape fullResGrid;

    private Texture2D rgbTexture;

    private Dictionary<int, Array> _activationTensorPerEpoch = new Dictionary<int, Array>();
    private int[] activationTensorShape = new int[4];

    public ImageLayer()
    {
    }

    protected override void UpdateForParams()
    {
        //float mult = 9f;
        //float res = mult + 1.5f + Mathf.Sin(Time.fixedTime)*mult;
        //reducedResolution.x = Mathf.FloorToInt(res);
        //reducedResolution.y = Mathf.FloorToInt(res);


        if (depth != oldDepth ||
            reducedResolution != oldreducedResolution
            || oldFullResolution != fullResolution
            || _initialized == false
            || _featureMaps == null)
        {
            InitFeatureMaps();
            oldreducedResolution = reducedResolution;
            oldFullResolution = fullResolution;
            oldDepth = depth;
        }

        if (fullResGrid == null)
            fullResGrid = new GridShape(new Vector3(0, 0, this.ZPosition() - fullresOffset), DrawMode.Dot, fullResolution, new Vector2(pixelSpacing, pixelSpacing));

        fullResGrid.position = new Vector3(0, 0, -fullresOffset);
        fullResGrid.spacing = new Vector2(pixelSpacing, pixelSpacing);
        fullResGrid.resolution = fullResolution;

        if (rgbTexture == null) { 
            rgbTexture = new Texture2D(3, 1);
            Color[] colors = { new Color(1f, 0, 0), new Color(0, 1f, 0), new Color(0, 0, 1f) };
            rgbTexture.SetPixels(colors);
            rgbTexture.filterMode = FilterMode.Point;
        }

        int totalVerts = _featureMaps.Count * reducedResolution.x * reducedResolution.y;
        _renderer.sharedMaterials[0].SetTexture("_MainTex", rgbTexture);
        _renderer.sharedMaterials[0].SetInt("_totalVerts", totalVerts);
    }

    public override void SetExpansion(float expansion)
    {

    }

    private void InitFeatureMaps()
    {
        int reducedDepthGridShape = Mathf.CeilToInt(Mathf.Sqrt(depth));
        Vector3[] xLinePositions = LineShape.ScaledUnitLine(depth, new Vector3(0, 0, 0), new Vector3(1, 0, 0), spread);
        Vector3[] zLinePositions = LineShape.ScaledUnitLine(depth, new Vector3(0, 0, 0), new Vector3(0, 0, 1), spread);

        Vector3[] filterPositions = Shape.InterpolateShapes(xLinePositions, zLinePositions, xToZ);

        _featureMaps = new List<FeatureMap>();
        for (int i = 0; i < depth; i++)
        {
            FeatureMap map = new FeatureMap(filterPositions[i], reducedResolution, currentConvShape, Get2DOutputShape(), pixelSpacing);
            _featureMaps.Add(map);
        }
    }

    private void UpdateFeatureMaps()
    {
        if(_featureMaps == null)
        {
            InitFeatureMaps();
            return;
        }
        int reducedDepthGridShape = Mathf.CeilToInt(Mathf.Sqrt(depth));
        Vector3[] xLinePositions = LineShape.ScaledUnitLine(depth, new Vector3(0, 0, 0), new Vector3(1, 0, 0), spread);
        Vector3[] zLinePositions = LineShape.ScaledUnitLine(depth, new Vector3(0, 0, 0), new Vector3(0, 0, 1), spread);

        Vector3[] filterPositions = Shape.InterpolateShapes(xLinePositions, zLinePositions, xToZ);

        for (int i = 0; i < _featureMaps.Count; i++)
        {
            _featureMaps[i].UpdateValues(filterPositions[i], reducedResolution, currentConvShape, _featureMaps[i].outputShape, pixelSpacing);
        }
    }

    public override void CalcMesh()
    {
        _mesh.subMeshCount = 3;

        List<Vector3> verts = new List<Vector3>();
        List<int> inds = new List<int>();
        List<int> fullResInds = new List<int>();
        List<int> linesInds = new List<int>();
        List<Color> cols = new List<Color>();

        AddNodes(verts, inds, fullResInds, linesInds);

        for (int i = 0; i < verts.Count; i++)
        {
            if (_activationTensorPerEpoch.ContainsKey(epoch))
            {

                int[] index = GetMultiDimIndices(activationTensorShape, i);

                Array activationTensor = _activationTensorPerEpoch[epoch];
                float tensorVal = (float)activationTensor.GetValue(index) * pointBrightness;
                cols.Add(new Color(tensorVal, tensorVal, tensorVal, 1f));

                //cols.Add(Color.white); 
            }
            else
            {
                cols.Add(Color.black);
            }
        }

        _mesh.SetVertices(verts);
        _mesh.SetColors(cols);
        _mesh.SetIndices(inds.ToArray(), MeshTopology.Points, 0);
        _mesh.SetIndices(fullResInds.ToArray(), MeshTopology.Points, 1);
        _mesh.SetIndices(linesInds.ToArray(), MeshTopology.Lines, 2);
    }

    protected void AddNodes(List<Vector3> verts, List<int> inds, List<int> fullResInds, List<int> lineInds)
    {
        base.AddNodes(verts, inds);

        if (showOriginalResolution)
        {
            Vector3[] v = fullResGrid.GetVertices(true);

            verts.AddRange(v);
            for (int i = 0; i < v.Length; i++)
            {
                fullResInds.Add(verts.Count - v.Length + i);
            }

            GridShape centerGrid = _featureMaps[_featureMaps.Count / 2].GetPixelGrid();
            float[] bbox = centerGrid.Bbox();

            float[] zpos = { 0, -fullresOffset };

            List<int> lineStartEndInds = new List<int>();
            foreach (float z in zpos)
            {
                verts.Add(new Vector3(bbox[0], bbox[1], centerGrid.position.z + z));
                lineStartEndInds.Add(verts.Count - 1);

                verts.Add(new Vector3(bbox[0], bbox[3], centerGrid.position.z + z));
                lineStartEndInds.Add(verts.Count - 1);

                verts.Add(new Vector3(bbox[2], bbox[1], centerGrid.position.z + z));
                lineStartEndInds.Add(verts.Count - 1);

                verts.Add(new Vector3(bbox[2], bbox[3], centerGrid.position.z + z));
                lineStartEndInds.Add(verts.Count - 1);
            }
            for (int i = 0; i < 4; i++)
            {
                lineInds.Add(lineStartEndInds[i]);
                lineInds.Add(lineStartEndInds[i + 4]);
            }
        }
    }

    protected override List<Shape> GetPointShapes()
    {
        UpdateFeatureMaps();

        List<Shape> pixelGrids = new List<Shape>();
        for (int i = 0; i < _featureMaps.Count; i++)
        {
            pixelGrids.Add(_featureMaps[i].GetPixelGrid());
        }
        return pixelGrids;
    }

    public override List<List<Shape>> GetLineStartShapes(Vector2Int convShape, Vector2Int outputShape, Vector2 theoreticalOutputShape, Vector2Int stride, float allCalcs)
    {
        currentConvShape = convShape;
        UpdateFeatureMaps();

        List<List<Shape>> filterGrids = new List<List<Shape>>();
        for (int i = 0; i < _featureMaps.Count; i++)
        {
            filterGrids.Add(_featureMaps[i].GetFilterGrids(outputShape, theoreticalOutputShape, stride, allCalcs));
        }
        return filterGrids;
    }

    public override Vector3Int GetOutputShape()
    {
        return new Vector3Int(reducedResolution.x, reducedResolution.y, depth);
    }

    public override Vector2Int Get2DOutputShape()
    {
        return reducedResolution;
    }

    public void SetActivationTensorForEpoch(Array tensor, int epoch)
    {
        _activationTensorPerEpoch.Add(epoch, tensor);
    }

    public Array GetActivationTensorForEpoch(int epoch)
    {
        return _activationTensorPerEpoch[epoch];
    }

    public void SetActivationTensorShape(int[] tensorShape)
    {
        this.activationTensorShape = tensorShape;
    }

    public int[] GetActivationTensorShape()
    {
        return activationTensorShape;
    }
}
